package top.ninwoo.app.dfs.client.service;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.ResponseEntity;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;
import top.ninwoo.bishe.starter.service.NetworkService;
import top.ninwoo.commonapi.po.FileSlice;
import top.ninwoo.commonapi.po.MergeResult;
import top.ninwoo.commonapi.po.SplitResult;
import top.ninwoo.commonapi.service.TransferService;

import javax.annotation.PostConstruct;
import java.io.*;
import java.util.*;

/**
 * 文件服务
 */
@Service
public class FileService {

    @Value("${dfs.directory.output}")
    private String directoryPath;

    @Value("${dfs.directory.output}")
    private String directoryOutput;

    @Qualifier(value="fileRestTemplate")
    @Autowired
    private RestTemplate restTemplate;

    @Autowired
    private NetworkService networkService;

    @Autowired
    private TransferService transferService;

    Long clusterId = 11112l;
    String[] wxName_list = new String[]{"sate2","sate3","sate4","sate5","sate6","sate7","sate8","sate9","sate10","sate11",
                                       "sate12","sate13","sate14","sate15","sate16","sate17","sate18","sate19","sate20","sate21",
                                       "sate22","sate23","sate24","sate25","sate26","sate27","sate28","sate29","sate30"};//获取卫星名字

    private String[] ipList = null;


    @PostConstruct
    public void init() {
        //ipList = getIpList(11112l,"dfs_server");
        for (int k = 0; k < wxName_list.length; k++){
            ipList = getIpList(clusterId,wxName_list[k]);//通过卫星名字获取卫星ip
        }

    }

    public String[] getIpList(Long clusterId, String appName) {

        //List<String>  ipListS = networkService.getIpListByAppName(clusterId, appName);//用于存储容器ip的list集合

        List<String> ipListS = new ArrayList<>();
        for (int k = 1; k < wxName_list.length; k++){
            ipListS.add(networkService.getIpListByAppName(clusterId, wxName_list[k]).toString()) ;
        }
        String[] containerIp = new String[ipListS.size()];//将集合中的元素存在数组里，因为下边用的是数组
        for (int i = 0; i < ipListS.size(); i++) {
            if(!ipListS.isEmpty()){
                String ip_tmp = ipListS.get(i);
                String rep = ip_tmp.replaceAll("[\\[\\]]","");
                String[] split_list = rep.split("/");
                containerIp[i] = split_list[0]+":8084";
            }
        }
        return containerIp;
    }


    // 从文件中读取数据
    public byte[] readFile(String fileName) {
        FileInputStream fis = null;
        ByteArrayOutputStream bos = null;

        try {
            File file = new File(fileName);
            fis = new FileInputStream(file);
            bos = new ByteArrayOutputStream();
            byte[] b = new byte[1024 * 10];//存取读到的多个字节
            int len;//每次读取的有效字节的个数
            while ((len = fis.read(b)) != -1) {
                bos.write(b, 0, len);
            }
            return bos.toByteArray();
        } catch (Exception e) {
            if(fis != null) {
                try {
                    fis.close();
                } catch (IOException ioException) {
                    ioException.printStackTrace();
                }
            }
            if(bos != null) {
                try {
                    bos.close();
                } catch (IOException ioException) {
                    ioException.printStackTrace();
                }
            }
            throw new RuntimeException("文件读取错误" + e.getMessage());
        }
    }

    /**
     * 发送文件到分布式节点
     * @param fileId
     * @param fileBytes
     * @return
     */
    @Async("taskExecutor")
    public String sendFile(Long fileId, byte[] fileBytes) {
        // 通过集群服务器接口获取当前集群的节点数量
        //ipList = this.getIpList(clusterId,"dfs_server");

        //读入图片数据
        //byte[] fileBytes = readFile("1.jpg");

        for (int i = 0; i < wxName_list.length; i++){
            ipList = getIpList(clusterId,wxName_list[i]);//通过卫星名字获取卫星ip
        }

        int sliceNum = ipList.length; //根据卫星节点个数确定冗余后的数据碎片个数
        int origNum = 6;  //自己设定的一个图片被分割的原始数据的个数
        // todo 这里需要制定文件的fileId
        SplitResult splitResult = transferService.fileSplit(fileBytes, sliceNum, origNum);

        //顺序，依次发送碎片到各个服务器
        int ipIndex = 0;
        for (FileSlice slice : splitResult.getFileSliceList()) {
            slice.setFileId(fileId);
            while(true){
                try {
                    sendFileSlice(ipList[ipIndex], slice);
                    ipIndex = (ipIndex + 1) % ipList.length;
                    break;
                }catch (Exception e){
                    ipIndex = (ipIndex + 1) % ipList.length;
                }
            }

        }
        return "success";

    }

    // 发送文件切片到目标地址
    public String sendFileSlice(String targetIp, FileSlice fileSlice) {
        ResponseEntity<String> response = restTemplate.postForEntity("http://" + targetIp + "/data/put/", fileSlice, String.class);
        if (!response.getStatusCode().is2xxSuccessful()) {
            return "failed!";
        }

        return response.getBody();
    }

    public Set<FileSlice> getFileSlice(String targetIp, Long fileId) {
        FileSlice[] fileSlices = restTemplate.getForObject("http://" + targetIp + "/data/get/" + fileId, FileSlice[].class);
        HashSet<FileSlice> fileSliceSet = new HashSet<>();
        fileSliceSet.addAll(Arrays.asList(fileSlices));
        return fileSliceSet;

    }

    public String getFile(String fileName, Long fileId) {

       //ipList = this.getIpList(clusterId,"dfs_server");
        for (int i = 0; i < wxName_list.length; i++){
            ipList = getIpList(clusterId,wxName_list[i]);//通过卫星名字获取卫星ip
        }
        int sliceNum = ipList.length;
        //int sliceNum = 30;
        int origNum = 6;
        Set<FileSlice> result = new HashSet<>();

        //异常处理
        for (int index = 0; index < ipList.length; index++) {
            try {
                Set<FileSlice> fileSliceSet = getFileSlice(ipList[index], fileId);
                result.addAll(fileSliceSet);
            }catch (Exception e){

            }
            if(result.size() >= origNum) {
                break;
            }
        }
        //System.out.println("result:"+result.size());

        /*for (String ip : ipList) {
            Set<FileSlice> fileSliceSet = getFileSlice(ip, fileId);
            result.addAll(fileSliceSet);
            if(result.size() >= origNum) {
                break;
            }
        }*/
        if(result.size() < origNum) {
            throw new RuntimeException("不可恢复的文件");
        }
        SplitResult splitResult = new SplitResult();
        splitResult.setFileSliceList(new ArrayList<>(result));

        MergeResult mergeResult = transferService.sliceMerge(splitResult, sliceNum, origNum);
        try {
            saveFile(fileName, mergeResult);
            return "success";
        } catch (Exception e) {
            e.printStackTrace();
            return "fail";
        }
    }

    public void saveFile(String fileName, MergeResult mergeResult) throws Exception {
        byte[] fileByte = mergeResult.getFileBytes();

        //将字节数组转为文件----可以验证读取图片到byte数组是否正确
        FileOutputStream fos = new FileOutputStream(directoryOutput + fileName);
        /*FileOutputStream fos = new FileOutputStream("F:/恢复.PNG");*/
        fos.write(fileByte, 0, fileByte.length);
        fos.close();
    }

}
