package top.ninwoo.app.dfs.service;

import top.ninwoo.app.dfs.util.Fraction;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Arrays;

public class EncodeAndDecode{

    //编码过程
    /*1、将传进来的byte数组拆分，也即原数组等大小拆分，同时转为BigInteger型(1*m)，同时将其转为m*1型，以便进行数据冗余*/
    public static BigInteger[][] fileByteSplit(byte[] data,int m){
        //int m = 7; //切割的片数
        int subSize = data.length % m == 0 ? data.length / m : data.length / m + 1;  //切割的数组大小
        BigInteger[] fileByteSplitToBigInt = new BigInteger[m];//fileByteSplitToBigInt数组用于存放分割后的数组
        for (int i = 0; i < m; i++) {
            int from = (int) (i * subSize);
            int to = (int) (from + subSize);
            fileByteSplitToBigInt[i] = new BigInteger(1, Arrays.copyOfRange(data, from, to));
        }
        //fileByteSplitToBigInt数组的转置-----filebytes(m*1)型
        BigInteger[][] filebytesBigInt = new BigInteger[fileByteSplitToBigInt.length][1];
        int num = 0;
        for(int i = 0; i<fileByteSplitToBigInt.length; i++){
            for(int j = 0;j<1; j++){
                filebytesBigInt[i][j] = fileByteSplitToBigInt[num];
                num++;
            }
        }
        return filebytesBigInt;
    }
    /*2 将filebytes(BigInteger型)转为BigDecimal型*/
    public static BigDecimal[][] bigIntToDecimal(BigInteger[][] bigInt){
        BigDecimal[][] filebytesBigDec = new BigDecimal[bigInt.length][bigInt[0].length];
        for (int i = 0; i < filebytesBigDec.length; i++) {
            for (int j = 0; j < filebytesBigDec[0].length; j++) {
                filebytesBigDec[i][j] = new BigDecimal(bigInt[i][j]);
            }
        }
        return filebytesBigDec;
    }
    /*3、创建编码矩阵---上半部分的单位矩阵+下半部分的范德蒙矩阵*/
    public static BigDecimal[][] encodeMatrix(int m, int shards){
        BigDecimal[][] encodeMatrixDec = new BigDecimal[m+shards][m];
        for (int i = 0; i < m; i++) {
            for (int j = 0; j < encodeMatrixDec[0].length; j++) {
                if(i == j){
                    encodeMatrixDec[i][j] = BigDecimal.valueOf(1);
                }else{
                    encodeMatrixDec[i][j] = BigDecimal.valueOf(0);
                }
            }
        }
        for (int i = m; i < encodeMatrixDec.length; i++) {
            for (int j = 0; j < encodeMatrixDec[0].length; j++) {
                encodeMatrixDec[i][j] = BigDecimal.valueOf(Math.pow(i-m+1,j));
            }
        }
        return encodeMatrixDec;
    }
    /*4、两数组相乘，生成编码数据块encodeDataMatrix（BigDecimal型数组(m+shards)*1）*/
    public static BigDecimal[][] encodeData(Fraction[][] a, BigDecimal[][] b){
        BigDecimal[][] encodeDataMatrixBigDec = new BigDecimal[a.length][b[0].length];
        for (int i = 0 ; i < encodeDataMatrixBigDec.length ; i++ ){
            for(int j = 0 ; j < encodeDataMatrixBigDec[0].length ; j++ ){//正好为矩阵大小，保证了每个元素都被赋值
                encodeDataMatrixBigDec[i][j] = BigDecimal.valueOf(0);
                for (int k = 0 ; k < b.length ; k++){
                    BigDecimal tmp = a[i][k].isNegate() ? a[i][k].getMole().multiply(b[k][j]).divide(a[i][k].getDeno(),0).negate()
                            : a[i][k].getMole().multiply(b[k][j]).divide(a[i][k].getDeno(),0);
                    encodeDataMatrixBigDec[i][j] = encodeDataMatrixBigDec[i][j].add(tmp);
                }
            }
        }
        return encodeDataMatrixBigDec;
    }

    public static BigDecimal[][] encodeData(BigDecimal[][] a, BigDecimal[][] b){
        BigDecimal[][] encodeDataMatrixBigDec = new BigDecimal[a.length][b[0].length];
        for (int i = 0 ; i < encodeDataMatrixBigDec.length ; i++ ){
            for(int j = 0 ; j < encodeDataMatrixBigDec[0].length ; j++ ){//正好为矩阵大小，保证了每个元素都被赋值
                encodeDataMatrixBigDec[i][j] = BigDecimal.valueOf(0);
                for (int k = 0 ; k < b.length ; k++){
                    encodeDataMatrixBigDec[i][j] = encodeDataMatrixBigDec[i][j].add(a[i][k].multiply(b[k][j]));
                }
            }
        }
        return encodeDataMatrixBigDec;
    }
    /*BigDecimal型数组转为BigInteger数组，并且将其转置为(1*m)*/
    public static BigInteger[] bigDecToInt(BigDecimal[][] bigDec){
        //将BigDecimal转为BigInteger
        BigInteger[][] encodeDataMatrixBigInt = new BigInteger[bigDec.length][bigDec[0].length];
        for (int i = 0; i < encodeDataMatrixBigInt.length; i++) {
            for (int j = 0; j < encodeDataMatrixBigInt[0].length; j++) {
                bigDec[i][j] = bigDec[i][j].setScale(0, BigDecimal.ROUND_HALF_UP);
                encodeDataMatrixBigInt[i][j] = bigDec[i][j].toBigInteger();
            }
        }
        //将BigInteger数组转置
        BigInteger[] encodeDataMatrixTran = new BigInteger[encodeDataMatrixBigInt.length];
        for (int i = 0; i < encodeDataMatrixBigInt.length; i++) {
            for (int j = 0; j < encodeDataMatrixBigInt[0].length; j++) {
                encodeDataMatrixTran[i+j] = encodeDataMatrixBigInt[i][j];
                //System.out.println(Arrays.toString(encodeDataMatrixTran[i + j].toByteArray()));
            }
        }
        return encodeDataMatrixTran;
    }

    /*-----------------------求逆矩阵---------------------------*/
    /*
     * 1、求(h,v)坐标的位置的余子式
     */
    public BigDecimal[][] getConfactor(BigDecimal[][] data, int h, int v) {
        int H = data.length;
        int V = data[0].length;
        BigDecimal[][] newdata = new BigDecimal[H-1][V-1];
        for(int i=0; i<newdata.length; i++) {
            if(i < h-1) {
                for(int j=0; j<newdata[i].length; j++) {
                    if(j < v-1) {
                        newdata[i][j] = data[i][j];
                    }else {
                        newdata[i][j] = data[i][j+1];
                    }
                }
            }else {
                for(int j=0; j<newdata[i].length; j++) {
                    if(j < v-1) {
                        newdata[i][j] = data[i+1][j];
                    }else {
                        newdata[i][j] = data[i+1][j+1];
                    }
                }
            }
        }
        return newdata;
    }
    /*
     * 2、计算行列式的值
     */
    public BigDecimal getMartrixResult(BigDecimal[][] data) {
        /*
         * 二维矩阵计算
         */
        if(data.length == 2) {
            return data[0][0].multiply(data[1][1]).subtract(data[0][1].multiply(data[1][0]));
        }
        /*
         * 二维以上的矩阵计算
         */
        BigDecimal result = BigDecimal.valueOf(0);
        int num = data.length;
        BigDecimal[] nums = new BigDecimal[num];
        for(int i=0; i<data.length; i++) {
            if(i%2 == 0) {
                nums[i] = data[0][i].multiply(getMartrixResult(getConfactor(data, 1, i+1)));
            }else {
                nums[i] = data[0][i].multiply(getMartrixResult(getConfactor(data, 1, i+1))).negate();
            }
        }
        for(int i=0; i<data.length; i++) {
            result = result.add(nums[i]);
        }
//      System.out.println(result);
        return result;
    }
    /*实现通过求代数余子式并转置来实现求逆矩阵*/
    public Fraction[][] getReverseMartrix(BigDecimal[][] data) {
        Fraction[][] newdata = new Fraction[data.length][data[0].length];
        BigDecimal A = getMartrixResult(data);
//      System.out.println(A);
        for(int i=0; i<data.length; i++) {
            for(int j=0; j<data[0].length; j++) {
                if((i+j)%2 == 0) {
                    newdata[i][j] = Fraction.builder().mole(getMartrixResult(getConfactor(data, i+1, j+1))).deno(A).build();
                }else {
                    newdata[i][j] = Fraction.builder().mole(getMartrixResult(getConfactor(data, i+1, j+1))).deno(A).negate(true).build();
                }
            }
        }
        newdata = trans(newdata);

/*        for(int i=0;i<newdata.length; i++) {
            for(int j=0; j<newdata[0].length; j++) {
                System.out.print(newdata[i][j]+ "   ");
            }
            System.out.println();
        }*/
        return newdata;
    }

    private Fraction[][] trans(Fraction[][] newdata) {
        // TODO Auto-generated method stub
        Fraction[][] newdata2 = new Fraction[newdata[0].length][newdata.length];
        for(int i=0; i<newdata.length; i++)
            for(int j=0; j<newdata[0].length; j++) {
                newdata2[j][i] = newdata[i][j];
            }
        return newdata2;
    }

    /*BigInteger数组转为byte数组*/
    public static byte[] bigIntegerToBytes(BigInteger[] bigIntegers,int length){
        //1、 origDataBigInt型数组转为byte数组[[byte1],[byte2],[byte3],...]
        byte[] array = null;
        byte[][] origDataBigIntToByte = new byte[bigIntegers.length][];
        for (int j = 0; j < bigIntegers.length; j++) {
            //System.out.println("byteList:"+byteList.get(j));
            array = bigIntegers[j].toByteArray();
            if (array[0] == 0) {
                byte[] tmp1 = new byte[array.length - 1];
                System.arraycopy(array, 1, tmp1, 0, tmp1.length);
                array = tmp1;
            }
            origDataBigIntToByte[j] = array;
        }

        //2、再将origDataBigIntToByte数组中的多个小数组合并，组成数组fileByteDest
        int length_byte = 0; //数组fileByteDest的长度
        for (byte[] bytes : origDataBigIntToByte) {
            length_byte += bytes.length;
        }
        // 转成的数组
        byte[] fileByteDestB = new byte[length_byte];
        int countLength = 0;
        for (byte[] arr : origDataBigIntToByte) {
            System.arraycopy(arr, 0, fileByteDestB, countLength, arr.length);
            countLength += arr.length;
        }
        //重新定义一个byte数组
        byte[] fileByteDest = new byte[length];
        System.arraycopy(fileByteDestB, 0, fileByteDest, 0, length);

        return fileByteDest;
    }

}
