package top.ninwoo.app.dfs.service;


import org.springframework.stereotype.Service;
import top.ninwoo.app.dfs.util.Fraction;
import top.ninwoo.commonapi.po.FileSlice;
import top.ninwoo.commonapi.po.MergeResult;
import top.ninwoo.commonapi.po.SendResult;
import top.ninwoo.commonapi.po.SplitResult;
import top.ninwoo.commonapi.service.TransferService;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;


@Service
public class FileTransferServiceImpl implements TransferService {

    @Override
    public SplitResult fileSplit(byte[] fileBytes, int sliceNum, int origNum) {

        EncodeAndDecode encode = new EncodeAndDecode();

        /*1、将传进来的byte数组拆分，也即原数组等大小拆分，同时转为BigInteger型，并将其转置成m*1型*/
        BigInteger[][] filebytesBigInt =  encode.fileByteSplit(fileBytes,origNum);
        //encode.printMartixInt(filebytesBigInt);

        /*2 将filebytes(BigInteger型)转为BigDecimal型*/
        BigDecimal[][] filebytesBigDec = encode.bigIntToDecimal(filebytesBigInt);

        /*3、创建编码矩阵---上半部分的单位矩阵+下半部分的范德蒙矩阵*/
        BigDecimal[][] encodeDecimal = encode.encodeMatrix(origNum,sliceNum-origNum);

        /*4、生成编码数据块encodeDataMatrix（BigInteger型数组(m+shards)*1）*/
        //两个BigInteger型数组相乘
        BigDecimal[][] encodeDataMatrixBigDec = encode.encodeData(encodeDecimal,filebytesBigDec);

        /*5、将BigDecimal转为BigInteger*/
        BigInteger[] encodeDataMatrixTran = encode.bigDecToInt(encodeDataMatrixBigDec);

        /*6、将数据碎片数组包装成方法返回值类型*/
        List<FileSlice> fileSliceList = new ArrayList<>();
        for (int j = 0; j < encodeDataMatrixTran.length; j++) {
            FileSlice fileSlice = new FileSlice();
            fileSlice.setFileId(1L);//文件的唯一标识
            fileSlice.setSliceId(j);//数据碎片的id
            fileSlice.setSize(fileBytes.length);//文件比特数据的有效长度
            fileSlice.setFileBytes(encodeDataMatrixTran[j].toByteArray());//文件的比特数据
            fileSliceList.add(fileSlice);
        }
        SplitResult splitResult = new SplitResult();
        splitResult.setFileSliceList(fileSliceList);
        return splitResult;//这是最后要返回的东西，也就是图片分割的结果
    }

    @Override
    public MergeResult sliceMerge(SplitResult splitResult, int sliceNum, int origNum) {

        EncodeAndDecode encode = new EncodeAndDecode();


        List<FileSlice> fileSliceListInput = splitResult.getFileSliceList();
        BigInteger[][] encodeDataBigInt = new BigInteger[fileSliceListInput.size()][1]; //从fileSliceList中任取m个数据碎片---7
        for (int i = 0; i < fileSliceListInput.size(); i++) {
            encodeDataBigInt[i][0] = new BigInteger(fileSliceListInput.get(i).getFileBytes());
        }

        //encodeDataBigInt转为encodeDataBigDec
        BigDecimal[][] encodeDataBigDec = encode.bigIntToDecimal(encodeDataBigInt);

        /*从编码矩阵中取对应的行，形成新的encodeMatrixNew*/
        BigDecimal[][] encodeMatrix = encode.encodeMatrix(origNum,sliceNum-origNum);

        BigDecimal[][] encodeMatrixNew = new BigDecimal[fileSliceListInput.size()][fileSliceListInput.size()];
        for (int i = 0; i < encodeMatrixNew.length; i++) {
            for (int j = 0; j < encodeMatrixNew[0].length; j++) {
                encodeMatrixNew[i][j] = encodeMatrix[fileSliceListInput.get(i).getSliceId()][j];
            }
        }
        /*求新的编码矩阵的逆矩阵*/
        Fraction[][] encodeMatrixNewInv = encode.getReverseMartrix(encodeMatrixNew);

        /*两矩阵相乘*/
        BigDecimal[][] origDataBigDec = encode.encodeData(encodeMatrixNewInv,encodeDataBigDec);

        /*BigDecimal转BigInteger*/
        BigInteger[] origDataBigInt = encode.bigDecToInt(origDataBigDec);

        /*origDataBigInt型数组转为byte数组(origDataBytes)*/
        byte[] fileByteDest = encode.bigIntegerToBytes(origDataBigInt,fileSliceListInput.get(0).getSize());

        MergeResult mergeResult = new MergeResult();
        mergeResult.setFileBytes(fileByteDest);

        return mergeResult;
    }

    @Override
    public SendResult sendFileSlice(String targetIp, FileSlice fileSlice) {
        return null;
    }

    @Override
    public FileSlice getFileSlice(String targetIp, Long fileId, int sliceId) {
        return null;
    }



}
